# Generated by the protocol buffer compiler.  DO NOT EDIT! - Oops. I did...
# source: MalwarePatternType.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
# from google.protobuf import symbol_database as _symbol_database  # had to comment this one because it doesnt exist in the protobuf py3 port
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

# _sym_db = _symbol_database.Default()  #  see why this is commented above




DESCRIPTOR = _descriptor.FileDescriptor(
  name='MalwarePatternType.proto',
  package='',
  serialized_pb=_b('\n\x18MalwarePatternType.proto\"{\n\x12MalwarePatternType\x12\x36\n\x0cpattern_type\x18\x01 \x02(\x0e\x32 .MalwarePatternType.PATTERN_TYPE\"-\n\x0cPATTERN_TYPE\x12\x0b\n\x07LANDING\x10\x01\x12\x10\n\x0c\x44ISTRIBUTION\x10\x02')
)
# _sym_db.RegisterFileDescriptor(DESCRIPTOR)



_MALWAREPATTERNTYPE_PATTERN_TYPE = _descriptor.EnumDescriptor(
  name='PATTERN_TYPE',
  full_name='MalwarePatternType.PATTERN_TYPE',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LANDING', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DISTRIBUTION', index=1, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=106,
  serialized_end=151,
)
# _sym_db.RegisterEnumDescriptor(_MALWAREPATTERNTYPE_PATTERN_TYPE)


_MALWAREPATTERNTYPE = _descriptor.Descriptor(
  name='MalwarePatternType',
  full_name='MalwarePatternType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pattern_type', full_name='MalwarePatternType.pattern_type', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _MALWAREPATTERNTYPE_PATTERN_TYPE,
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=28,
  serialized_end=151,
)

_MALWAREPATTERNTYPE.fields_by_name['pattern_type'].enum_type = _MALWAREPATTERNTYPE_PATTERN_TYPE
_MALWAREPATTERNTYPE_PATTERN_TYPE.containing_type = _MALWAREPATTERNTYPE
DESCRIPTOR.message_types_by_name['MalwarePatternType'] = _MALWAREPATTERNTYPE

MalwarePatternType = _reflection.GeneratedProtocolMessageType('MalwarePatternType', (_message.Message,), dict(
  DESCRIPTOR = _MALWAREPATTERNTYPE,
  __module__ = 'MalwarePatternType_pb2'
  # @@protoc_insertion_point(class_scope:MalwarePatternType)
  ))
# _sym_db.RegisterMessage(MalwarePatternType)


# @@protoc_insertion_point(module_scope)
